<?php

/*
  continue-flightplan: V 1.0 2023 by Thomas Breitenbücher, breitenbuecher.de

  command prompt e.g.: php -f %USERPROFILE%\Desktop\FS-Tools\Tools\develop\sinkrate\continue-flightplan.php [switch-aircraft]
  my web environment:  http://127.0.0.1/continue-flightplan.php
*/

/*
  Functions:
  - checkParameter ()                                 check parameters e.g. for aircraft switching
  - getCustomFlightPlan ()                            extract flightplan-data out of customflight.pln as array
  - getNameOfNewFlightFile ()                         get name of flt-file to write new data
  - checkCrashedSimulator ()                          the simulator might crash after loading a flt-file
  - checkFlightPlanIntegrity ()                       simple checks if the flight plan data might be ok
  - removeTimeWaypoints ()                            remove TIME-Waypoints from new ftl-file and customflight.pln because the simulator might crash when loading a modified flt-file with reduced waypoint-data
  - checkSimStates ()                                 check for sim states (active flightplan and sim on ground)
  - backupFiles ()                                    make copies of the customflight-flt/pln-files and the the saved flt-file
  - getCurrentAirportICAO ()                          get ICAO-Code of current airport
  - setDepartureICAOCode ()                           set ICAO-code in departure-section to current airport
  - reduceOldWaypoints ()                             delete old waypoints and renumber entries of several sections
  - reduceOldWaypointsIcaoIndexInCustomFlightPlan ()  find index-number of given ICAO-code in customflight.pln as double-check
  - reduceOldWaypointsFindNearestWaypointIndex ()     helper function for finding nearest waypoint index
  - reduceOldWaypointsOfCustomFlightSections ()       helper function for reducing old waypoints in the customflight-sections
  - reduceOldWaypointsCopyFlightSection ()            helper function for copying reduced flightplan-entries from 'ATC_Aircraft.0' into 'ATC_ActiveFlightPlan.0'
  - reduceOldWaypointsOfCustomFlightPlanSections ()   helper function for reducing flightplan-entries in customflight.pln
  - setFreeFlightState ()                             set FreeFlightState to a suitable value
  - importSavedFlightStates ()                        import saved flight states into a array-copy of customflight.flt
  - cleanFlightPlanData ()                            clean flight-plan-data when a flight-plan has been finished
  - restoreLostFlightPlan ()                          restore existing customflight.pln-data after user-prompt if flightplan-data is lost in both flt-files
  - writeFlightFile ()                                write array data to flt-file
  - writeMessage ()                                   write a message according to environment
  - getStringFromFlightFileArray ()                   converts customflight-array-data into a microsoft-flt-file-string
  - criticalError ()                                  exit showing error information according to environment
  - getUserinput ($s)                                 give prompt message and wait for user input
  - switchAircraft ()                                 switch a choosen aircraft to use in the recently saved flight
  - switchCrashOptions ()                             switch an option after a detecting crash
  - switchCrashOptionFltFile ()                       switch an crash option for the flt-file
  geolocation:
  - getDecimalFromCoordinateString ()                 wrapper for transforming coordinate-strings in to signed coordinate-values
  - getArrayFromCoordinateString ()                   gives 3 array values for a coordinate string like N48° 57' 30.63"
  - DegreesMinutesSeconds2Decimal ()                  convert geolocation data from degrees°-minutes"-seconds'-format into decimal format
  - Decimal2DegreesMinutesSeconds ()                  convert geolocation data from decimal format into degrees°-minutes"-seconds'-format (not used here)
  - getDistanceBetweenCoordinates ()                  calculates the distance between decimal-formatted-coordinates in km, miles or nautical miles
  - getAirportDataFromLnmApi ()                       get airport data from JSON-data of LittleNavMap-API (coordinates and length of longest runway)
*/

/*
  common information:
  - waypoint format:
      region:   region code (e.g. K1),
      ident:    ICAO ident,
      airport:  name of airport or real waypoint-id
      name:     name of airport or simulator-specific waypoints or 'TIMECLIMB' etc.
  - basic thoughts:
    - basic flt-flightplan-data will be read from customflight.flt and modified in $arrCustomFlight before being transferred into TargetFltFile and customflight.pln
*/

// CONSTANTS

// user constants
define ('DEBUG', false);                                                      // Debug: don't write backup- and new flt-file, but show array-data
define ('CONTINUE_DATE_TIME', false);                                         // true: start next flight with date & time of the saved flight; false: start with current time (deleting date&season-section)
define ('USE_LITTLENAVMAP_API', true);                                        // true: connect a running server instance of LittleNavMap, false: skip connection

// MSFS files
define ('FLIGHT_DATA_DIR', getenv('USERPROFILE') . '\AppData\Local\Packages\Microsoft.FlightSimulator_8wekyb3d8bbwe\LocalState' . "\\");
define ('CUSTOM_FLIGHT_DIR', FLIGHT_DATA_DIR . 'MISSIONS\Custom\CustomFlight' . "\\");
define ('CUSTOM_FLIGHT', CUSTOM_FLIGHT_DIR . 'CustomFlight.FLT');
define ('CUSTOM_FLIGHTPLAN', CUSTOM_FLIGHT_DIR . 'CustomFlight.PLN');

// MSFS special-files
define ('START_FS_REFERENCE_FILE', 'PublishingGroups.json');                  // to detect new flt-files
define ('TEMPORARY_FLIGHT_FILE', 'TemporaryFlight.Flt');                      // to detect an intermediate temporary flight
define ('LOCK_FILE', FLIGHT_DATA_DIR . 'running.lock');                       // lockfile, should be deleted with normal shutdown of simulator

// various constants
define ('SEPARATOR_BLANK', ' ');
define ('SEPARATOR_COMMA', ',');
define ('SEPARATOR_PERIOD', ".");
define ('SEPARATOR_MINUTES', "'");
define ('SEPARATOR_SECONDS', '"');
define ('SEPARATOR_MINUS_SIGN', "-");
define ('SEPARATOR_EQUAL_SIGN', "=");
define ('FREEFLIGHT_STATE_IN_AIR', 'FLIGHT_CRUISE');                          // FLIGHT_CRUISE gives a good result in most in-air-cases
define ('FREEFLIGHT_STATE_ON_GROUND', 'PREFLIGHT_GATE');                      // FLIGHT_RUNWAY is temporary (in case of trouble a crash-option may be offered to change the value of FirstFlightState, PREFLIGHT_GATE should give a good result in most on-ground-cases, PREFLIGHT_HOLDSHORT, LANDING_GATE and FLIGHT_RUNWAY are alternatives
define ('LONGEST_RUNWAY_LENGTH_OF_SMALL_AIRPORT', 4500);                      // longest runway length in feet of a small airport to tell apart small airports from big ones and to tell later about a outback situation
define ('DISTANCE_FROM_BIG_AIRPORT_FOR_BEING_OUTBACK', 3);                    // difference in nm from nearest airport to rate an outback parking (if airport is big, decimal-'.'-values are allowed)
define ('DISTANCE_FROM_SMALL_AIRPORT_FOR_BEING_OUTBACK', 1);                  // difference in nm from nearest airport to rate an outback parking (if airport is small, decimal-'.'-values are allowed)

define ('URL_LNM_API', 'http://localhost:8965/api/airport/info?ident=');      // URL for retrieving airport data from LittleNavMap in JSON-format

define ('SWITCH_AIRCRAFT_INI', dirname (__FILE__) . '\aircraft-switch.ini');  // ini-file containing aircraft-strings according the ones saved in flt-Sim.0-sim or the aircraft-selector-develop-mode-list
define ('PARAMETER_SWITCH_AIRCRAFT', 'switch-aircraft');                      // script parameter (via command line or ?-URL) for switching aircrafts

// GLOBAL VARIABLES

$bSimOnGround = false;                                                        // is sim on ground?
$bSwitchAircraft = false;                                                     // alternative mode "switch-aircraft"?
$bActiveFlightPlan = false;                                                   // is there an active flight plan?
$bFlightPlanFinished = false;                                                 // has a flight plan been finished?
$iNumberOfRestoredGpsWaypoints = false;                                       // was a flight plan restored because of missing data?

$sErrors = '';
$sTargetFltFile = '';                                                         // path and name of newest flt-file (later used as target file with the modified data of $arrCustomFlight)
$sCustomFlightPlan = '';                                                      // custom-flightplan can't be read into an array (will be re-written with modified flight-plan-data)
$sCustomFlightData = '';                                                      // buffer for modified custom-flight-data to be written in new-flight-file
$sCurrentAirportICAO = '';                                                    // ICAO-code of the current airport where the flight-plan-execution has stopped

$arrFltFiles = array();                                                       // list of flt-files with modification time stamps
$arrCustomFlight = array ();                                                  // custom-flight as array (with modifications as content for the target-flight-file)
$arrNewFlightFile = array ();                                                 // newest flt-file as array

// CODE

writeMessage ('', true);

checkParameter ();

if (DEBUG == true)
  writeMessage ('Debug mode');

// read file-content
$arrCustomFlight = parse_ini_file (CUSTOM_FLIGHT, true, INI_SCANNER_RAW);
$sCustomFlightPlan = file_get_contents (CUSTOM_FLIGHTPLAN);
$arrCustomFlightPlan = getCustomFlightPlan ();                                // Ex-Parameter: $sCustomFlightPlan

checkCrashedSimulator ();     // crash after loading a flt-file?

// determine new flt-file
$sTargetFltFile = FLIGHT_DATA_DIR . getNameOfNewFlightFile ();
$arrNewFlightFile = parse_ini_file ($sTargetFltFile, true, INI_SCANNER_RAW);

checkFlightPlanIntegrity ();  // flight plan data might be ok
removeTimeWaypoints ();       // clean all TIME-waypoints
checkSimStates ();            // set global vars $bSimOnGround and $bActiveFlightPlan and start restoring flight-plan-data if necessary

// make backups from relevant files
backupFiles ($sTargetFltFile, $bActiveFlightPlan, $bSwitchAircraft);

// determine location, reduce waypoints, set and import states
$sCurrentAirportICAO = getCurrentAirportICAO ();
setDepartureICAOCode ($sCurrentAirportICAO);
reduceOldWaypoints ($sCurrentAirportICAO);
setFreeFlightState ($sCurrentAirportICAO);
importSavedFlightStates ($sCurrentAirportICAO);

// delete flight-plan-data when a flight-plan has been finished
if ($bFlightPlanFinished == true)
  cleanFlightPlanData ();

// write files

$sCustomFlightData = getStringFromFlightFileArray ($arrCustomFlight);

writeFlightFile ($sCustomFlightData, $sTargetFltFile);

if ($bActiveFlightPlan == true)
  writeFlightFile ($sCustomFlightPlan, CUSTOM_FLIGHTPLAN);

if ($bFlightPlanFinished == true)
  writeFlightFile ($sCustomFlightData, CUSTOM_FLIGHT);

if (DEBUG == true)
{
  echo '<pre>';
  echo '<hr />';
  echo '<h4>final debug</h4>';
  echo '<p>$sCustomFlightData to be written in: ' . $sTargetFltFile . '</p>';
  echo '<p>' . $sCustomFlightData . '</p>';
//  echo '<hr />';
//  echo '<p>$arrNewFlightFile</p>';
//  print_r ($arrNewFlightFile);
//  echo '<hr />';
//  echo '<p>$arrCustomFlight</p>';
//  print_r ($arrCustomFlight);
  echo '<hr />';
  echo '<p>$sCustomFlightPlan</p>';
  echo htmlentities ($sCustomFlightPlan);
  echo '</pre>';
}

// stop execution
exit;


// FUNCTIONS


function checkParameter ()
{
  global $bSwitchAircraft;

  $sParam = '';

  if (isset ($_SERVER['argv']))
  {
    if (isset ($_SERVER['argv'][1]))
      $sParam = $_SERVER['argv'][1];
  }
  elseif (count ($_GET) == 1)
    $sParam = key ($_GET);

  if ($sParam == PARAMETER_SWITCH_AIRCRAFT)
  {
    $bSwitchAircraft = true;
    switchAircraft ();
    exit;
  }
}


function getCustomFlightPlan ()
{
  global $sCustomFlightPlan;

  // the customflight.pln-data couldn't be read into an array, so we fiddle with string-functions to create array-data

  // catching flightplan part

  if ($sCustomFlightPlan == '')
  {
    writeMessage ('no flight plan data');
    return array ();
  }

/*
  $iPos1 = stripos ($sCustomFlightPlan, '<ATCWaypoint');                                  // start of customflight.pln-data
  $iPos2 = strripos ($sCustomFlightPlan, '</ATCWaypoint>') + strlen ('</ATCWaypoint>');   // end of customflight.pln-data

  $sExtractedFlightPlanData = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);
*/

// REMOVE TIME-ENTRIES
// Umstellung: statt $sExtractedFlightPlanData wird hier verwiesen auf $sCustomFlightPlan
  $iPos1 = 0;
  $iPos2 = strrpos ($sCustomFlightPlan, '</ATCWaypoint>') + strlen ('</ATCWaypoint>');         // last occurance
  $iLenOfTail = strlen ($sCustomFlightPlan) - $iPos2;                   // strlen ($sCustomFlightPlan);
//die ('<p>LenOfTail: ' . $iLenOfTail . '</p>');

/*
echo '<hr />';
echo '<p>before: $sCustomFlightPlan</p>';
echo '<pre>!' . htmlentities ($sCustomFlightPlan) . '!</pre>';
echo '<hr />';
*/
  // loop through all string elements for removing TIME-entries
  while ($iPos1 < strlen ($sCustomFlightPlan) - $iLenOfTail)     // $iLenOfEntries
  {
    $iPos1 = stripos ($sCustomFlightPlan, '<ATCWaypoint', $iPos1);
    $iPos2 = stripos ($sCustomFlightPlan, '</ATCWaypoint>', $iPos1) + strlen ('</ATCWaypoint>');
    $sXmlContentOflightPlanEntry = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1 + 10);

  if (stripos ($sXmlContentOflightPlanEntry, 'id="TIME'))
  {
//echo '<p>found: $sXmlContentOflightPlanEntry ' . strlen ($sXmlContentOflightPlanEntry) . '</p>';
//echo '<pre>found: !' . htmlentities ($sXmlContentOflightPlanEntry) . '!</pre>';
 $sCustomFlightPlan = str_replace ($sXmlContentOflightPlanEntry, '', $sCustomFlightPlan);
 $iLenOfEntries = strlen ($sCustomFlightPlan);
//echo '<p>removed: $sCustomFlightPlan</p>';
//echo '<pre>!' . htmlentities ($sCustomFlightPlan) . '!</pre>';
//exit;
//    $iPos1 = $iPos2;
  }
  else
  {
    $iPos1 = $iPos2;
  }


//echo '<p>$iPos1 => $iPos2: '. $iPos1 . ' => ' . $iPos2 . '</p>';
  }

/*
echo '<hr />';
echo '<p>after: $sCustomFlightPlan</p>';
echo '<pre>!' . htmlentities ($sCustomFlightPlan) . '!</pre>';
echo '<hr />';
*/

// war weiter oben
  $iPos1 = stripos ($sCustomFlightPlan, '<ATCWaypoint');                                  // start of customflight.pln-data
  $iPos2 = strripos ($sCustomFlightPlan, '</ATCWaypoint>') + strlen ('</ATCWaypoint>');   // end of customflight.pln-data

  $sExtractedFlightPlanData = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  // $sExtractedFlightPlanData contains the ATCWaypoint-structure-entries: identify all entries and put them into an array

  $iPos1 = 0;
  $iPos2 = 0;
  $iLenOfEntries = strlen ($sExtractedFlightPlanData);

  // find whitespace-string between the ATCWaypoint-structure-entries for correct reassembling afterwards
  $iPos1 = stripos ($sExtractedFlightPlanData, '</ATCWaypoint>') + strlen ('</ATCWaypoint>');
  $iPos2 = stripos ($sExtractedFlightPlanData, '<ATCWaypoint', $iPos1);
  $sCustomFlightplanWhiteSpace = substr ($sExtractedFlightPlanData, $iPos1, $iPos2 - $iPos1);

//  $iPos = 0;
  $iPos1 = 0;
  $iPos2 = 0;
  $arr = array ();
  $arrTempIndex = array ();
  $arrTempValues = array ();
  $arrCustomFlightPlan = array ();
  $arrCustomFlightPlanEntries = array ();

  // loop through all string elements filling the array $arrCustomFlightPlanEntries
  while ($iPos1 < $iLenOfEntries)
  {
    $arrCustomFlightPlanEntries = array ();
    $iPos1 = stripos ($sExtractedFlightPlanData, '<ATCWaypoint', $iPos1);
    $iPos2 = stripos ($sExtractedFlightPlanData, '</ATCWaypoint>', $iPos1) + strlen ('</ATCWaypoint>');
    $sXmlContentOflightPlanEntry = substr ($sExtractedFlightPlanData, $iPos1, $iPos2 - $iPos1);

    $xmlParser = xml_parser_create ();
    xml_parse_into_struct ($xmlParser, $sXmlContentOflightPlanEntry, $arrTempValues, $arrTempIndex);
    xml_parser_free ($xmlParser);

    // transfer xml-array-data into MS-like array structure with numeric indizes starting with 0

    foreach ($arrTempIndex as $xmlKey => $xmlIndices)
    {
      // $xmlKey: text of data-element
      // $xmlIndices[0]: index-number in $arrTempValues (for the corresponding array)

      switch ($xmlKey)
      {
        case 'ATCWAYPOINT':     $arr = array ($xmlKey => $arrTempValues[$xmlIndices[0]]['attributes']['ID']);
                                break;

        case 'ATCWAYPOINTTYPE': $arr = array ($xmlKey => $arrTempValues[$xmlIndices[0]]['value']);
                                break;

        case 'WORLDPOSITION':   $arr = array ($xmlKey => $arrTempValues[$xmlIndices[0]]['value']);
                                break;

        case 'SPEEDMAXFP':      $arr = array ($xmlKey => $arrTempValues[$xmlIndices[0]]['value']);
                                break;

        case 'ICAOIDENT':       $arr = array ($xmlKey => $arrTempValues[$xmlIndices[0]]['value']);
                                break;
      }

      // append new entries
      $arrCustomFlightPlanEntries = array_merge ($arrCustomFlightPlanEntries, $arr);
    }

//    $iPos++;
    $iPos1 = $iPos2;
    $arrCustomFlightPlan[] = $arrCustomFlightPlanEntries;

  }

  return $arrCustomFlightPlan;
}

function getNameOfNewFlightFile ($bOutput = true, $bCrash = false)
{
  global $bSwitchAircraft, $arrFltFiles;

  $sTimestampRef = '';
  $dir = new DirectoryIterator (FLIGHT_DATA_DIR);

  // list flt-files in flight-data-directory
  foreach ($dir as $fileinfo)
  {
    $sFileName = $fileinfo->getFilename();

    if (stripos ($sFileName, '.flt') && stripos ($sFileName, 'backup') === false)
      $arrFltFiles[$fileinfo->getMTime()] = $sFileName;

    if ($sFileName == START_FS_REFERENCE_FILE)
      $sTimestampRef = $fileinfo->getMTime();
  }

  // sort descending
  krsort ($arrFltFiles);

  // now the first array entry points to the newest flt-file
  $sFileName = current ($arrFltFiles);

  // check, if the newest flt-file is newer than last start of simulator
  $TimeStamp = key ($arrFltFiles);

  if ($TimeStamp > $sTimestampRef)
  {
    if ($bOutput == true)
      writeMessage ('found new flight file ' . $sFileName);
  }
  else
  {
    if ($bSwitchAircraft == false)
    {
      // exit only in standard mode
      if ($bCrash == false)
      {
        writeMessage ('no new flight file');
        exit;
      }
    }
    else
      writeMessage ('no new flight, choose another');
  }

  if ($sFileName == TEMPORARY_FLIGHT_FILE)
  {
    if ($bSwitchAircraft == false)
    {
      // exit only in standard mode
      writeMessage ('temporary flight, no further action');
      exit;
    }
    else
      writeMessage ('temporary flight, choose another');
  }

  return $sFileName;
}


function checkCrashedSimulator ()
{
  if (file_exists (LOCK_FILE))
  {
    writeMessage ("crash after loading a flt-file?");
    switchCrashOptions ();
  }
}


function checkFlightPlanIntegrity ()
{
  global $arrNewFlightFile;

  // Rewriting a flight file with integer flight-plan-data might corrupt the flight file.
  // Therefor a simple check might do some minimal action.

  if ($arrNewFlightFile['ATC_Aircraft.0']['ActiveFlightPlan'] == 'True')
  {
    // a flight plan should exist in the new flight file
    // so we expect at least corresponding waypoint.0-entries in the sections ATC_Aircraft.0 and ATC_ActiveFlightPlan.0
    if (array_key_exists ('Waypoint.0', $arrNewFlightFile['ATC_Aircraft.0']))
    {
      if (array_key_exists ('ATC_ActiveFlightPlan.0', $arrNewFlightFile))
      {
        if (array_key_exists ('Waypoint.0', $arrNewFlightFile['ATC_ActiveFlightPlan.0']))
        {
          // we believe in an integer flight-plan-data-structure and do only some action

          writeMessage ('flight plan data should be ok');

          checkSimStates (true);
          $sCurrentAirportICAO = getCurrentAirportICAO ();
          setFreeFlightState ($sCurrentAirportICAO, true);
          $sCustomFlightData = getStringFromFlightFileArray ($arrNewFlightFile);
          writeFlightFile ($sCustomFlightData, CUSTOM_FLIGHT);

          exit;
        }
      }
    }
  }

  // in all other cases the show must go on
}


function removeTimeWaypoints ()
{
  global $arrCustomFlight, $arrCustomFlightPlan;

  // at this point the ATC_ActiveFlightPlan.0-section doesn't contain flight plan data
  // so we deal with the ATC_Aircraft.0- and GPS_Engine-section of $arrCustomFlight and the customflight.pln-data

  $iWaypointIndex = 0;

/*
if (DEBUG == true)
{
  echo '<pre>';
  echo '<hr />';
  echo '<h4>temp develop: TIME-Waypoints entfernen (alle Sektionen in flt; auch pln)</h4>';
  echo '<p>ATC_Aircraft.0</p>';
  print_r ($arrCustomFlight['ATC_Aircraft.0']);
  echo '<p>ATC_ActiveFlightPlan.0</p>';
  print_r ($arrCustomFlight['ATC_ActiveFlightPlan.0']);
  echo '<p>GPS_Engine</p>';
  print_r ($arrCustomFlight['GPS_Engine']);
}
*/

  foreach ($arrCustomFlight['ATC_Aircraft.0'] as $key => $val)
  {
    if (stripos ($key, 'waypoint.') === 0)
    {
      if (strpos ($val, ', , , TIME') === 0)
      {
//        echo $key . ': ' . $val . ' => ' . $iWaypointIndex . '<br />';
        unset ($arrCustomFlight['ATC_Aircraft.0'][$key]);
        unset ($arrCustomFlight['GPS_Engine']['WpInfo' . $iWaypointIndex]);
      }
      
      $iWaypointIndex++;
    }
  }

/*
if (DEBUG == true)
{
  echo '<p>removed: ATC_Aircraft.0</p>';
  print_r ($arrCustomFlight['ATC_Aircraft.0']);
  echo '<p>removed: GPS_Engine</p>';
  print_r ($arrCustomFlight['GPS_Engine']);
  echo '<hr />';
  echo '<h4>jetzt noch customflight.pln erleichtern und nachsehen, ob auch ein restoreFlightplan mit den bereinigten Daten klar kommt ...</h4>';
  echo '<p>$arrCustomFlightPlan</p>';
  print_r ($arrCustomFlightPlan);
}
*/

// DAS MUSS WOANDERS PASSIEREN MIT DEM STRING
/*
  foreach ($arrCustomFlightPlan as $key => $val)
  {
    if (strpos ($val['ATCWAYPOINT'], 'TIME') === 0)
    {
      unset ($arrCustomFlightPlan[$key]);
    }
  }

if (DEBUG == true)
{
  echo '<p>$arrCustomFlightPlan</p>';
  print_r ($arrCustomFlightPlan);
}
*/

}


function checkSimStates ($bFlightPlanDataShouldBeOk = false)
{
  global $arrCustomFlight, $arrNewFlightFile, $sCustomFlightPlan, $bSimOnGround, $bActiveFlightPlan;

  $s = '';
  $iTimeStampPln = 0;
  $iTimestampRef = 0;

  // is sim on ground?

  if (strtoupper ($arrNewFlightFile['SimVars.0']['SimOnGround']) == 'TRUE')
    $bSimOnGround = true;

  // no further action if flight plan data should be ok when called by checkFlightPlanIntegrity ()

  if ($bFlightPlanDataShouldBeOk == true)
  {
    $bActiveFlightPlan = true;
    return;
  }

  // Case 1: Is there an active flightplan in customflight.flt?
  // This is more reliable, because flight-plan-data is often lost in the saved flight-file.

  if (strtoupper ($arrCustomFlight['ATC_Aircraft.0']['ActiveFlightPlan']) == 'TRUE')
    $bActiveFlightPlan = true;

  // Case 2: Sometimes even in customflight.flt the connection to an existing flightplan might get lost.

  // If there's no-flight-plan-data in customflight.flt, but flight plan data in customflight.pln,
  // then we'll try to restore the flight-plan-data without asking the user.

  // so restoring flight-plan-data is done in the above mentioned cases

  if (($bActiveFlightPlan == true) || ($bActiveFlightPlan == false && strpos ($sCustomFlightPlan, '<FlightPlan.FlightPlan>')))
  {
    // don't look on customflight.pln-file if it's older than last simulator-start

    $iTimestampPln = filemtime (CUSTOM_FLIGHTPLAN);
    $iTimestampRef = filemtime (FLIGHT_DATA_DIR . START_FS_REFERENCE_FILE);

    if ($iTimestampRef > $iTimestampPln)
      writeMessage ('not looking at customflight.pln');
    else
      $bActiveFlightPlan = restoreLostFlightPlan ();
  }

  if (DEBUG == true)
    writeMessage ('Debug: $bSimOnGround=' . $bSimOnGround . ', $bActiveFlightPlan=' . $bActiveFlightPlan);
}


function backupFiles ($sTargetFltFile, $bActiveFlightPlan, $bSwitchAircraft)
{
  // php-copy will overwrite existing files

  $sError = '';
  $sSuffix = '.backup';

  if (DEBUG == false)
  {
    if ($bSwitchAircraft == true)
    {
      if (!copy (CUSTOM_FLIGHT, CUSTOM_FLIGHT . $sSuffix))
        $sError .= 'customflight ';

      return;
    }

    if (!copy ($sTargetFltFile, $sTargetFltFile . $sSuffix))
      $sError .= 'targetfile ';

    if ($bActiveFlightPlan == true)
    {
      if (!copy (CUSTOM_FLIGHT, CUSTOM_FLIGHT . $sSuffix))
        $sError .= 'customflight ';

      if (!copy (CUSTOM_FLIGHTPLAN, CUSTOM_FLIGHTPLAN . $sSuffix))
        $sError .= 'customflightplan';
    }

    if ($sError == '')
      writeMessage ('making backups');
    else
      writeMessage ('error making backups: ' . $sError);
  }
  else
    writeMessage ('Debug: no backups in debug-mode, $bSwitchAircraft:' . $bSwitchAircraft . ', $bActiveFlightPlan:' . $bActiveFlightPlan);
}


function getCurrentAirportICAO ()
{
  global $arrNewFlightFile, $arrCustomFlight, $bActiveFlightPlan, $bSimOnGround;

  $sICAO = '';
  $sMessage = '';
  $sDistance1 = '';
  $sSmallOrBig = '';
  $iDistance = 0;
  $iDistToCompare = 0;
  $arrSelection = array ();
  $arrAirportData = array ();

  // Unfortunately the departure-section is not reliable or the flt-file doesn't contain waypoints.
  // Sometimes the departure-section doesn't exist or points to the inital airport, sometimes it points to the airport of the last stop.

  // In a saved flights the new ICAO-code is always stored as current airport (based on the [SimVars.0]-values of [Latitude] and [Longitude]).
  // The distances will be calculated, stored in an array and after array-sorting the nearest airport can be identified.

  // Differences between the flight states are only made in setFreeFlightState ().

  $sLatitude  = $arrNewFlightFile['SimVars.0']['Latitude'];
  $sLongitude = $arrNewFlightFile['SimVars.0']['Longitude'];

  $arrLatitude = getArrayFromCoordinateString ($sLatitude);
  $arrLongitude = getArrayFromCoordinateString ($sLongitude);

  $decimalLatitude = DegreesMinutesSeconds2Decimal ($arrLatitude[0], $arrLatitude[1], $arrLatitude[2]);
  $decimalLongitude = DegreesMinutesSeconds2Decimal ($arrLongitude[0], $arrLongitude[1], $arrLongitude[2]);

  if ($bActiveFlightPlan == true)
    $sMessage = 'found flightplan, ';
  else
    $sMessage = 'no flightplan, ';

  if (isset ($arrCustomFlight['ATC_Aircraft.0']['Waypoint.0']))
  {
    // with existing flightplan-waypoints
    foreach ($arrCustomFlight['ATC_Aircraft.0'] as $key => $val)
    {
      if (stripos ($key, "waypoint.") === 0)
      {
        $arr = explode (SEPARATOR_COMMA, $val);

        // the second position contains the ICAO code or whitespace
        $sICAO = trim ($arr[1]);

        if ($arr[0] == '' && $sICAO != '')
        {
          $iDistance = getDistanceBetweenCoordinates ($decimalLatitude, $decimalLongitude, getDecimalFromCoordinateString ($arr[5]), getDecimalFromCoordinateString ($arr[6]));
          $arrSelection[$sICAO] = $iDistance;    // the distance between ICAO-code-location and coordinates is stored as value
        }
      }
    }
  }
  else
  {
    // no flightplan
    if ($bActiveFlightPlan == false)
    {
      if (isset ($arrCustomFlight['Departure']['ICAO']))
      {
        // but a ICAO-value given in departure-section
        $sICAO = $arrCustomFlight['Departure']['ICAO'];                       // read ICAO-value from departure-section
        $arrAirportData = getAirportDataFromLnmApi ($sICAO);                  // retrieving data from a running LittleNavMap-Server-API

        if (count ($arrAirportData) > 0)                                      // we are expecting array entries with decimal values for latitude and longitude
        {
          $iDistance = getDistanceBetweenCoordinates ($decimalLatitude, $decimalLongitude, $arrAirportData['lat'], $arrAirportData['lon']);
          $arrSelection[$sICAO] = $iDistance;                                  // the distance between ICAO-code-location and coordinates is stored as value
        }
      }
      else
        writeMessage ('no flightplan and no ICAO code');
    }
    else
      writeMessage ('flightplan with no waypoint entries');
  }

  if (count ($arrSelection) > 0)
  {
    asort ($arrSelection);
    $sICAO = key ($arrSelection);
    $sDistance = round ($arrSelection[$sICAO], 1) . ' nm';

    if (count ($arrAirportData) == 0)
      $arrAirportData = getAirportDataFromLnmApi ($sICAO);                    // retrieving data from a running LittleNavMap-Server-API

    if (count ($arrAirportData) > 0)
    {
      // setting values for $sSmallOrBig and $iDistToCompare
      if ($arrAirportData['longestRunwayLength'] > LONGEST_RUNWAY_LENGTH_OF_SMALL_AIRPORT)
      {
        $sSmallOrBig = 'big';
        $iDistToCompare = DISTANCE_FROM_BIG_AIRPORT_FOR_BEING_OUTBACK;
      }
      else
      {
        $sSmallOrBig = 'small';
        $iDistToCompare = DISTANCE_FROM_SMALL_AIRPORT_FOR_BEING_OUTBACK;
      }

      if ($bSimOnGround == true)
      {
        $sMessage .= 'sim on ground ';

        if ($arrSelection[$sICAO] > $iDistToCompare)
          $sMessage .= 'in outback (' . $sDistance . ' from ' . $sSmallOrBig . ' ' . $sICAO . ')';
        else
          $sMessage .= '(' . $sDistance . ' inside ' . $sSmallOrBig . ' ' . $sICAO . ')';
      }
      else
        $sMessage = 'sim not on ground (' . $sDistance . ' from ' . $sSmallOrBig . ' ' . $sICAO . ')';

      writeMessage ($sMessage);
    }
  }
  elseif ($bActiveFlightPlan == true)
    writeMessage ('flightplan with no waypoint entries in ' . __FUNCTION__ . '()');

  // at least we should have an ICAO-code

  if (DEBUG == true)
    writeMessage ('Debug: ICAO-Code=' . $sICAO);

  if ($sICAO == '')
    criticalError ('no ICAO-Code found in: ' . __FUNCTION__ . '()');

  return $sICAO;
}


function setDepartureICAOCode ($sICAO)
{
  global $arrCustomFlight;

  // Only the ICAO-value is set.
  // This will display the current airport in the flight-preview.
  // It's ok to ignore other values which will contain wrong data afterwards (like 'GateNumber' or 'RunwayNumber').

  if (DEBUG == true)
    writeMessage ('Debug: to ' . $sICAO);

  $arrCustomFlight['Departure']['ICAO'] = $sICAO;
}


function reduceOldWaypoints ($sICAO)
{
  global $sCustomFlightPlan, $bActiveFlightPlan, $bFlightPlanFinished;

  $iCurrentWaypointIndex = 0;

  $arr = array ();
  $arrNewWaypoints = array ();

  if ($bActiveFlightPlan == false)
    return;

  if ($sICAO == '')
  {
    writeMessage ('no ICAO-code given');
    return;
  }

  // catching the case of unwanted repeated reducing, especially while developing
  if (reduceOldWaypointsIcaoIndexInCustomFlightPlan ($sICAO) == 0)
  {
    writeMessage ('nothing to reduce being at #0');
  }

  // get index of current waypoint

  $iCurrentWaypointIndex = reduceOldWaypointsFindNearestWaypointIndex ($sICAO);

  if ($bFlightPlanFinished == true)
    return;

  writeMessage ('reducing flt-flightplan-waypoints (current position: #' . $iCurrentWaypointIndex . ')');

  // loop through CustomFlight-entries to get content of
  // - ATC_Aircraft.0-section and
  // - GPS_Engine-section
  // to build arrays with
  // - data before waypoints
  // - a pure waypoint-array with numeric keys
  // - data after waypoints

  $arrNewWaypoints = reduceOldWaypointsOfCustomFlightSections ('ATC_Aircraft.0', 'Waypoint.', 'NumberofWaypoints', $iCurrentWaypointIndex);

/*
echo __FUNCTION__ . ': $iCurrentWaypointIndex=' . $iCurrentWaypointIndex . ', $arrNewWaypoints' . PHP_EOL;
print_r ($arrNewWaypoints);
*/

  reduceOldWaypointsOfCustomFlightSections ('GPS_Engine', 'WpInfo', 'CountWP', $iCurrentWaypointIndex);

  // customflight.flt also contains a 'ATC_ActiveFlightPlan.0'-section which simply can be replaced by a copy of $arrNewWaypoints

  reduceOldWaypointsCopyFlightSection ('ATC_ActiveFlightPlan.0', 'waypoint.', $arrNewWaypoints);

  // at least the customflight.pln-file (referenced in the new flight-file) contains flightplan-data which also have to be reduced

  // Reducing flt-flightplan-data with $iCurrentWaypointIndex is fine.
  // But inside the customflight.pln-data this might lead to errors, so this is based on the more reliable $sICAO.

  $sCustomFlightPlan = reduceOldWaypointsOfCustomFlightPlanSections ($sICAO);
}


function reduceOldWaypointsIcaoIndexInCustomFlightPlan ($sICAO)
{
  global $arrCustomFlightPlan;

  $i = -1;

  foreach ($arrCustomFlightPlan as $key => $value)
  {
    if (isset ($value['ICAOIDENT']))
    {
      if ($value['ICAOIDENT'] == $sICAO)
      {
        $i = $key;

        // maybe an airport was reached (1), from where a side trip was planned, which ends at the formerly reached airport (2)
        // 1. this airport is not #0 and former waypoints will be deleted
        // 2. this airport exists at #0 and #n and might be found when continuing

        if ($key == 0)
          continue;
        else
          break;
      }
    }
  }

/*
// ABSTECHER abchecken
echo __FUNCTION__ . ': $sICAO=' . $sICAO . ', #' . $i . PHP_EOL;
echo 'Abstecher abchecken...' . PHP_EOL;
echo '- davor gibts was zu Kürzen - ok' . PHP_EOL;
echo '- davor gibts nichts zu kürzen und $sICAO taucht nochmal auf - den nächsten nehmen' . PHP_EOL;
echo __FUNCTION__ . ' - $arrCustomFlightPlan: <pre>';
print_r ($arrCustomFlightPlan);
echo '</pre>';
*/
  return $i;
}


function reduceOldWaypointsFindNearestWaypointIndex ($sICAO)
{
  global $arrCustomFlight, $bActiveFlightPlan, $bFlightPlanFinished;

  $iWpElements = 0;
  $iWayPointIndex = 0;
  $bWayPointFound = false;
  $arrWayPointData = array ();

  // flightplan might be reduced in all cases of saved flights: parking at airport, in-air or parking-in-outback

  // first the nearest waypoint-index should be found (by ICAO-Code as #1 or by coordinates #5 and #6)

  // find waypoint-index of current airport

/*
// ABSTECHER abchecken
echo __FUNCTION__ . ' - $arrCustomFlight[ATC_Aircraft.0]: <pre>';
print_r ($arrCustomFlight['ATC_Aircraft.0']);
echo '</pre>';
*/

  foreach ($arrCustomFlight['ATC_Aircraft.0'] as $key => $val)
  {
    if (stripos ($key, 'waypoint.') === 0)
    {
      $arrWayPointData = explode (SEPARATOR_COMMA, $val);

      // the second position contains the ICAO code or whitespace
      $s = trim ($arrWayPointData[1]);

      // maybe an airport was reached, from where a side trip was planned -> then this airport is not #0
      // if this airport is reached again (after a side trip) -> then it's #0 and it will appear again later in the flt-flightplan
      // the loop is walked till the end to
      // - get the total numer of flt-waypoints to detect a final arrival (aka finished flightplan) later
      // - try to catch a following appearance of an side-trip-airport

      if ($s == $sICAO && $bWayPointFound == false && $iWpElements > 0)
      {
// echo 'Abstecher?: ' . $s . ', Elemente: ' . $iWpElements . '<br />';
        $bWayPointFound = true;
        $iWayPointIndex = $iWpElements;
      }

      $iWpElements++;
    }
  }

/*
  if ($bActiveFlightPlan == true && $iWayPointIndex == 0)
  {
    writeMessage ('?flightplan with no waypoints in ' . __FUNCTION__ . '()');
    return 0;
  }
*/

  // flightplan is finished when current waypoint meets the number of waypoints (aka: is last one)

  if ($iWayPointIndex == $iWpElements - 1)
  {
    $bFlightPlanFinished = true;
    writeMessage ('arrived at destination, flightplan finished');
//    exit;
  }

  return $iWayPointIndex;
}


function reduceOldWaypointsOfCustomFlightSections ($sCustomFlightSection, $sWaypointString, $sNameOfNumberEntry, $iCurrentWaypointIndex)
{
  global $iNumberOfRestoredGpsWaypoints, $sCurrentAirportICAO, $arrCustomFlight;

  // loop through waypoint-entries to build a pure waypoint-array with numeric keys
  // additionally 2 more arrays are created containing the data before or after the waypoint-entries

  $bWaypoint = false;
  $arrWaypoints = array ();
  $arrNewWaypoints = array ();
  $arrBeforeWaypoints = array ();
  $arrAfterWaypoints = array ();

  // corrections when a restore of flight-plan-data has occured before

  if ($sCustomFlightSection == 'GPS_Engine')
  {
    if ($iNumberOfRestoredGpsWaypoints == true)
    {
      $iCurrentWaypointIndex = $iNumberOfRestoredGpsWaypoints - $arrCustomFlight['ATC_Aircraft.0']['NumberofWaypoints'];
      $arrCustomFlight['ATC_Aircraft.0']['AcState'] = 'ACSTATE_VFR_ENROUTE';
      $arrCustomFlight['ATC_Aircraft.0']['ActiveVFRAirport'] = $sCurrentAirportICAO;
      $arrCustomFlight['ATC_Aircraft.0']['WaypointNext'] = 1;
      $arrCustomFlight['ATC_ActiveFlightPlan.0']['departure_id'] = $sCurrentAirportICAO;
      $arrCustomFlight['GPS_Engine']['NextWP'] = 1;
      $arrCustomFlight['GPS_Engine']['CountWP'] = $arrCustomFlight['ATC_Aircraft.0']['NumberofWaypoints'];
    }
  }

  // go on

  if (isset ($arrCustomFlight[$sCustomFlightSection]))
  {
    foreach ($arrCustomFlight[$sCustomFlightSection] as $key => $val)
    {
      if (stripos ($key, $sWaypointString) === 0)
      {
        if ($bWaypoint == false)
          $bWaypoint = true;

        $arrWaypoints[] = $val;
      }
      else
      {
        if ($bWaypoint == false)
          $arrBeforeWaypoints[$key] = $val;
        else
          $arrAfterWaypoints[$key] = $val;
      }
    }

    // rebuild array of reduced waypoints with correct keys

    if (count ($arrWaypoints) > 0)
    {
      $i = 0;
      foreach ($arrWaypoints as $key => $val)
      {
        if ($key >= $iCurrentWaypointIndex)
        {
          $arrNewWaypoints[$sWaypointString . $i] = $val;
          $i++;
        }
      }
    }
    else
      writeMessage ('no waypoints found');

    // now we have 3 parts to create the newly built $arrCustomFlight[$sCustomFlightSection]-section (containing reduced flightplan-data)

    $arrCustomFlight[$sCustomFlightSection] = array_merge ($arrBeforeWaypoints, $arrNewWaypoints, $arrAfterWaypoints);

    // in $arrCustomFlight[$sCustomFlightSection] we finally have to adjust the value of $sNameOfNumberEntry

    $arrCustomFlight[$sCustomFlightSection][$sNameOfNumberEntry] = count ($arrNewWaypoints);
  }
  else
    writeMessage ($sCustomFlightSection . ' will not be available!');

  return $arrNewWaypoints;
}


function reduceOldWaypointsCopyFlightSection ($sCustomFlightSection, $sWaypointString, $arrNewWaypoints)
{
  global $arrCustomFlight;

  // $arrCustomFlight also contains a 'ATC_ActiveFlightPlan.0'-section which simply can be replaced by 2 arrays

  $bWaypoint = false;
  $arrBeforeWaypoints = array ();

  if (isset ($arrCustomFlight[$sCustomFlightSection]))
  {
    foreach ($arrCustomFlight[$sCustomFlightSection] as $key => $val)
    {
      if (stripos ($key, $sWaypointString) === 0)
      {
        if ($bWaypoint == false)
          $bWaypoint = true;
      }
      else
      {
        if ($bWaypoint == false)
          $arrBeforeWaypoints[$key] = $val;
      }
    }

    // now we have 2 parts to create the newly built $arrCustomFlight[$sCustomFlightSection]-section (containing reduced flightplan-data from above)

    $arrCustomFlight[$sCustomFlightSection] = array_merge ($arrBeforeWaypoints, $arrNewWaypoints);
  }
}


function reduceOldWaypointsOfCustomFlightPlanSections ($sICAO)    // ($iCurrentWaypointIndex)
{
  global $sCustomFlightPlan, $arrCustomFlightPlan;

  $iCurrentWaypointIndex = 0;                                                             // dealing with a separate waypoint-index inside this function

/*
echo __FUNCTION__ . ':' . PHP_EOL;
echo $sCustomFlightPlan;
echo '$sICAO = ' . $sICAO . ', array $arrCustomFlightPlan:' . PHP_EOL;
print_r ($arrCustomFlightPlan);
*/

  // $iCurrentWaypointIndex is newly determined not relying on the former $iCurrentWaypointIndex (being correct for flt-flightplan-data)

  foreach ($arrCustomFlightPlan as $key => $val)
  {
    if (isset ($val['ICAOIDENT']))
    {
      if ($val['ICAOIDENT'] == $sICAO)
      {
        $iCurrentWaypointIndex = $key;
        break;
      }
    }
  }

  // now we have a $iCurrentWaypointIndex which should be correct for the customflight.pln-data

  writeMessage ('reducing pln-flightplan-waypoints (current position: #' . $iCurrentWaypointIndex . ')');

  // the customflight.pln-data couldn't be read into an array, so we fiddle with string-functions to create array-data

  // catching content parts

  $iPos1 = stripos ($sCustomFlightPlan, '<ATCWaypoint');                                  // start of customflight.pln-data
  $iPos2 = strripos ($sCustomFlightPlan, '</ATCWaypoint>') + strlen ('</ATCWaypoint>');   // end of customflight.pln-data

  $sBeforeExtractedFlightPlanData = substr ($sCustomFlightPlan, 0, $iPos1);
  $sExtractedFlightPlanData = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);
  $sAfterExtractedFlightPlanData = substr ($sCustomFlightPlan, $iPos2);

  // $sExtractedFlightPlanData contains the ATCWaypoint-structure-entries: identify all entries and put them into an array

  $iPos1 = 0;
  $iPos2 = 0;
  $iLenOfEntries = strlen ($sExtractedFlightPlanData);

  // find whitespace-string between the ATCWaypoint-structure-entries for correct reassembling afterwards
  $iPos1 = stripos ($sExtractedFlightPlanData, '</ATCWaypoint>') + strlen ('</ATCWaypoint>');
  $iPos2 = stripos ($sExtractedFlightPlanData, '<ATCWaypoint', $iPos1);
  $sCustomFlightplanWhiteSpace = substr ($sExtractedFlightPlanData, $iPos1, $iPos2 - $iPos1);

  $iPos1 = 0;
  $iPos2 = 0;
  $arrCustomFlightPlanEntries = array ();

  // loop through all string elements filling the array $arrCustomFlightPlanEntries
  while ($iPos1 < $iLenOfEntries)
  {
    $iPos1 = stripos ($sExtractedFlightPlanData, '<ATCWaypoint', $iPos1);
    $iPos2 = stripos ($sExtractedFlightPlanData, '</ATCWaypoint>', $iPos1) + strlen ('</ATCWaypoint>');
    $arrCustomFlightPlanEntries[] = substr ($sExtractedFlightPlanData, $iPos1, $iPos2 - $iPos1);
    $iPos1 = $iPos2;
  }

  // delete all elements of $arrCustomFlightPlan-Entries below $iCurrentWaypointIndex

  foreach ($arrCustomFlightPlanEntries as $key => $val)
  {
    if ($key < $iCurrentWaypointIndex)
      unset ($arrCustomFlightPlanEntries[$key]);
  }

  // create string containing the reduced $arrCustomFlightPlan-Entries

  $sNewCustomFlightPlanEntries = '';

  foreach ($arrCustomFlightPlanEntries as $key => $val)
  {
    $sNewCustomFlightPlanEntries .= $val . $sCustomFlightplanWhiteSpace;
  }

  // remove last $sCustomFlightplanWhiteSpace

  $sNewCustomFlightPlanEntries = substr ($sNewCustomFlightPlanEntries, 0, - strlen ($sCustomFlightplanWhiteSpace));

/*
echo '$iCurrentWaypointIndex = ' . $iCurrentWaypointIndex . PHP_EOL;
echo '$sNewCustomFlightPlanEntries = ' . $sNewCustomFlightPlanEntries . PHP_EOL;
echo 'exit';
exit;
*/

  // and return the combines parts of the new customflight.pln

  return $sBeforeExtractedFlightPlanData . $sNewCustomFlightPlanEntries . $sAfterExtractedFlightPlanData;
}


function setFreeFlightState ($sCurrentAirportICAO, $bExeption = false)
{
  global $arrCustomFlight, $arrNewFlightFile, $bSimOnGround;

  // When saving a flt-file often misleading values are stored.

  // These values are kept unchanged when in-air.
  // Being on-ground (at-airport or in-outback) means setting a preferred FirstFlightState.

  if ($bExeption == false)
  {
    if ($bSimOnGround == true)
      $arrCustomFlight['FreeFlight']['FirstFlightState'] = FREEFLIGHT_STATE_ON_GROUND;
    else
      $arrCustomFlight['FreeFlight']['FirstFlightState'] = FREEFLIGHT_STATE_IN_AIR;
  }
  else
  {
    // $bSimOnGround doesn't contain an actuel value at this point
    if (strtoupper ($arrNewFlightFile['SimVars.0']['SimOnGround']) == 'TRUE')
      $arrNewFlightFile['FreeFlight']['FirstFlightState'] = FREEFLIGHT_STATE_ON_GROUND;
    else
      $arrNewFlightFile['FreeFlight']['FirstFlightState'] = FREEFLIGHT_STATE_IN_AIR;
  }
}


function importSavedFlightStates ($sCurrentAirportICAO)
{
  global $iNumberOfRestoredGpsWaypoints, $arrCustomFlight, $arrNewFlightFile;

  // set some data or import some aircraft-related data from saved flt-file into copy of customflight.flt

  $arrCustomFlight['Options']['SaveOriginalFlightPlan'] = 'True';

  $arrCustomFlight['Sim.0'] = $arrNewFlightFile['Sim.0'];
  $arrCustomFlight['ResourcePath'] = $arrNewFlightFile['ResourcePath'];
  $arrCustomFlight['ObjectFile'] = $arrNewFlightFile['ObjectFile'];
  $arrCustomFlight['Loading'] = $arrNewFlightFile['Loading'];

  // ['DateTimeSeason'] may be imported (continue at saved data & time) or deleted (continue at current date & time)
  if (CONTINUE_DATE_TIME == true)
    $arrCustomFlight['DateTimeSeason'] = $arrNewFlightFile['DateTimeSeason'];
  else
    unset ($arrCustomFlight['DateTimeSeason']);

  $arrCustomFlight['SimVars.0'] = $arrNewFlightFile['SimVars.0'];
  $arrCustomFlight['Avionics.0'] = $arrNewFlightFile['Avionics.0'];
  $arrCustomFlight['SimVars.0'] = $arrNewFlightFile['SimVars.0'];
  $arrCustomFlight['Avionics.0'] = $arrNewFlightFile['Avionics.0'];
  $arrCustomFlight['Engine Parameters.1.0'] = $arrNewFlightFile['Engine Parameters.1.0'];

  if (array_key_exists ('Propeller.1.0', $arrNewFlightFile))
    $arrCustomFlight['Propeller.1.0'] = $arrNewFlightFile['Propeller.1.0'];

  if (array_key_exists ('Fuel.0', $arrNewFlightFile))
    $arrCustomFlight['Fuel.0'] = $arrNewFlightFile['Fuel.0'];

  $arrCustomFlight['Gauges.0'] = $arrNewFlightFile['Gauges.0'];
//  $arrCustomFlight['LocalVars.0'] = $arrNewFlightFile['LocalVars.0'];
  $arrCustomFlight['Switches.0'] = $arrNewFlightFile['Switches.0'];
  $arrCustomFlight['Electrical.0'] = $arrNewFlightFile['Electrical.0'];
  $arrCustomFlight['Systems.0'] = $arrNewFlightFile['Systems.0'];
  $arrCustomFlight['AutoPilot.0'] = $arrNewFlightFile['AutoPilot.0'];
  $arrCustomFlight['Payload.0'] = $arrNewFlightFile['Payload.0'];
  $arrCustomFlight['Controls.0'] = $arrNewFlightFile['Controls.0'];
  $arrCustomFlight['SimScheduler'] = $arrNewFlightFile['SimScheduler'];
}


function cleanFlightPlanData ()
{
  global $arrNewFlightFile, $arrCustomFlight, $sCustomFlightPlan;

  $i = 0;

  // $arrNewFlightFile

  $arrNewFlightFile['ATC_Aircraft.0']['ActiveFlightPlan'] = 'False';
  unset ($arrNewFlightFile['ATC_ActiveFlightPlan.0']);

  if (isset ($arrNewFlightFile['ATC_Aircraft.0']['Waypoint.0']))
  {
    // with existing flightplan-waypoints
    foreach ($arrNewFlightFile['ATC_Aircraft.0'] as $key => $val)
    {
      if (stripos ($key, 'Waypoint.') === 0)
      {
        unset ($arrNewFlightFile['ATC_Aircraft.0']['Waypoint.' . $i]);
        $i++;
      }
    }
  }

  $arrNewFlightFile['ATC_Aircraft.0']['NumberofWaypoints'] = '0';
  $arrNewFlightFile['ATC_Aircraft.0']['WaypointNext'] = '0';
// AcState=ACSTATE_NONE

  $i = 0;

  if (isset ($arrNewFlightFile['GPS_Engine']['WpInfo0']))
  {
    // with existing flightplan-waypoints
    foreach ($arrNewFlightFile['GPS_Engine'] as $key => $val)
    {
      if (stripos ($key, 'WpInfo') === 0)
      {
        unset ($arrNewFlightFile['GPS_Engine']['WpInfo' . $i]);
        $i++;
      }
    }
  }

  unset ($arrNewFlightFile['GPS_Engine']);

  // $arrCustomFlight

  $arrCustomFlight['ATC_Aircraft.0']['ActiveFlightPlan'] = 'False';
  unset ($arrCustomFlight['ATC_ActiveFlightPlan.0']);

  $i = 0;

  if (isset ($arrCustomFlight['ATC_Aircraft.0']['Waypoint.0']))
  {
    // with existing flightplan-waypoints
    foreach ($arrCustomFlight['ATC_Aircraft.0'] as $key => $val)
    {
      if (stripos ($key, 'Waypoint.') === 0)
      {
        unset ($arrCustomFlight['ATC_Aircraft.0']['Waypoint.' . $i]);
        $i++;
      }
    }
  }

  $arrCustomFlight['ATC_Aircraft.0']['NumberofWaypoints'] = '0';
  $arrCustomFlight['ATC_Aircraft.0']['WaypointNext'] = '0';

  $i = 0;

  if (isset ($arrCustomFlight['GPS_Engine']['WpInfo0']))
  {
    // with existing flightplan-waypoints
    foreach ($arrCustomFlight['GPS_Engine'] as $key => $val)
    {
      if (stripos ($key, 'WpInfo') === 0)
      {
        unset ($arrCustomFlight['GPS_Engine']['WpInfo' . $i]);
        $i++;
      }
    }
  }

  unset ($arrCustomFlight['GPS_Engine']);

  // $sCustomFlightPlan

  $sCustomFlightPlan = '';
}


function restoreLostFlightPlan ()
{
  global $iNumberOfRestoredGpsWaypoints, $sTargetFltFile, $sCustomFlightPlan, $arrCustomFlight, $arrCustomFlightPlan, $arrNewFlightFile;

  $i = 0;
  $iWP = 0;
  $iPos1 = 0;
  $iPos2 = 0;
  $sKey = '';
  $sVal = '';
  $bSplitpoint = false;
  $arrWorldPosition = array ();
  $arrNewAtcAircraft = array ();
  $arrBeforeCustomFlight = array ();
  $arrAfterCustomFlight = array ();
  $arrNewAtcActiveFlightPlan = array ();

  if (count ($arrCustomFlightPlan) == 0)
    criticalError ('empty array $arrCustomFlightPlan in: ' . __FUNCTION__ . '()');
  else
    writeMessage ('restoring flight-plan-data');

  // walk through $arrNewFlightFile to find 'ATC_Aircraft.0'-elements and fill a new array with restored flight-plan-data

  foreach ($arrNewFlightFile as $key => $val)
  {

    if ($key == 'ATC_Aircraft.0')
    {
      // walk through the 'ATC_Aircraft.0'-elements

      foreach ($val as $key2 => $val2)
      {
        // set ActiveFlightPlan
        if ($key2 == 'ActiveFlightPlan')
          $arrNewAtcAircraft['ActiveFlightPlan'] = 'True';

        // copy the next 2 elements
        $arrNewAtcAircraft['RequestedFlightPlan'] = $val['RequestedFlightPlan'];
        $arrNewAtcAircraft['AcState'] = $val['AcState'];

        // and create this one
        $arrNewAtcAircraft['ActiveVFRAirport'] = $arrCustomFlightPlan[0]['ICAOIDENT'];
      }

      // loop through customflight.pln-data to create the missing flight-flt-data-elements

      foreach ($arrCustomFlightPlan as $keyPln => $valPln)
      {
        $sKey = 'Waypoint.' . $i;

        foreach ($arrCustomFlightPlan[$keyPln] as $keyPln2 => $valPln2)
        {
          $sVal = ', ';
          
          if (isset ($valPln['ICAOIDENT']))
            $sVal .= $valPln['ICAOIDENT'] . ', ';
          else
            $sVal .= ', ';

          $sVal .= ', ';

          $sVal .= $valPln['ATCWAYPOINT'] . ', ';
          
          if (isset ($valPln['ICAOIDENT']))
            $sVal .= 'A, ';
          else
            $sVal .= 'U, ';

          $arrWorldPosition = explode (',', $valPln['WORLDPOSITION']);

          $sVal .= $arrWorldPosition[0] . ', ' . $arrWorldPosition[1] . ', ' . $arrWorldPosition[2] . ', , , , , , NONE, 0, 0, -1, 0, 0, ';
        }

        $i++;
        $arrNewAtcAircraft[$sKey] = $sVal;
      }

      // create this one
      $arrNewAtcAircraft['NumberofWaypoints'] = $i;

      // copy the next 2 elements
      $arrNewAtcAircraft['ClearanceFlags'] = $val['ClearanceFlags'];
      $arrNewAtcAircraft['CtCur'] = $val['CtCur'];

      // set this one
      $arrNewAtcAircraft['WaypointNext'] = 1;

      // copy the rest of the 'ATC_Aircraft.0'-elements
      $arrNewAtcAircraft['AltCleared'] = $val['AltCleared'];
      $arrNewAtcAircraft['HdgAssigned'] = $val['HdgAssigned'];
      $arrNewAtcAircraft['SquawkAssigned'] = $val['SquawkAssigned'];
      $arrNewAtcAircraft['LandingSequence'] = $val['LandingSequence'];
      $arrNewAtcAircraft['DepartureRequest'] = $val['DepartureRequest'];
      $arrNewAtcAircraft['ParkingRequest'] = $val['ParkingRequest'];
      $arrNewAtcAircraft['ParkingTypeRequest'] = $val['ParkingTypeRequest'];
      $arrNewAtcAircraft['PatternLeg'] = $val['PatternLeg'];
      $arrNewAtcAircraft['ApproachIndex'] = $val['ApproachIndex'];
      $arrNewAtcAircraft['ApproachTransitionIndex'] = $val['ApproachTransitionIndex'];
      $arrNewAtcAircraft['ApproachRequestIndex'] = $val['ApproachRequestIndex'];
      $arrNewAtcAircraft['ApproachTransitionRequestIndex'] = $val['ApproachTransitionRequestIndex'];
      $arrNewAtcAircraft['RunwayIndex'] = $val['RunwayIndex'];
      $arrNewAtcAircraft['RunwayRequestIndex'] = $val['RunwayRequestIndex'];
      $arrNewAtcAircraft['ParkingIndex'] = $val['ParkingIndex'];
      $arrNewAtcAircraft['FlightFollowing'] = $val['FlightFollowing'];
      $arrNewAtcAircraft['AirspaceTransition'] = $val['AirspaceTransition'];
      $arrNewAtcAircraft['BVAirspaceTransition'] = $val['BVAirspaceTransition'];
      $arrNewAtcAircraft['BVAirspaceTransitionReq'] = $val['BVAirspaceTransitionReq'];
      $arrNewAtcAircraft['TakeoffSequence'] = $val['TakeoffSequence'];
      $arrNewAtcAircraft['LandingSequenceNumber'] = $val['LandingSequenceNumber'];
      $arrNewAtcAircraft['TaxiRouteCurrent'] = $val['TaxiRouteCurrent'];
      $arrNewAtcAircraft['CruisingAltitude'] = $val['CruisingAltitude'];
      $arrNewAtcAircraft['RequestedCruisingAltitude'] = $val['RequestedCruisingAltitude'];
      $arrNewAtcAircraft['ExpectedAltitude'] = $val['ExpectedAltitude'];
      $arrNewAtcAircraft['AircraftSignature'] = $val['AircraftSignature'];
    }
  }

  if (DEBUG == true)
  {
  /*
    echo '<pre>';
    echo '<hr />';
    echo __FUNCTION__ .' ():';
    echo '<p>$arrNewAtcAircraft</p>';
    print_r ($arrNewAtcAircraft);

    echo '<p>$arrCustomFlight</p>';
    print_r ($arrCustomFlight);
    echo '<pre>';
  */
  }

  // overwrite the restored 'ATC_Aircraft.0'-data
  $arrCustomFlight['ATC_Aircraft.0'] = $arrNewAtcAircraft;

  // now we have restored flight-plan-data in $arrCustomFlight
  
  // walk through $arrCustomFlight to split parts for injecting the missing 'ATC_ActiveFlightPlan.0'-section

  foreach ($arrCustomFlight as $key => $val)
  {
    if (stripos ($key, 'ATC_AircraftManager') === 0)
    {
      if ($bSplitpoint == false)
      {
        $bSplitpoint = true;
        $arrAfterWaypoints[$key] = $val;
      }
    }
    else
    {
      if ($bSplitpoint == false)
        $arrBeforeWaypoints[$key] = $val;
      else
        $arrAfterWaypoints[$key] = $val;
    }
  }
  
  // creating elements of missing 'ATC_ActiveFlightPlan.0'-section in $arrNewAtcActiveFlightPlan

  $iPos1 = strpos ($sCustomFlightPlan, '<Title>') + strlen ('<Title>');
  $iPos2 = strpos ($sCustomFlightPlan, '</Title>', $iPos1);
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['title'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['description'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1) . ', restored by ' . __FUNCTION__ . '()';

  $iPos1 = strpos ($sCustomFlightPlan, '<FPType>') + strlen ('<FPType>');
  $iPos2 = strpos ($sCustomFlightPlan, '</FPType>', $iPos1);
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['type'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['routetype'] = 0;  // ?
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['cruising_altitude'] = -1;

  $iPos1 = strpos ($sCustomFlightPlan, '<DepartureID>') + strlen ('<DepartureID>');
  $iPos2 = strpos ($sCustomFlightPlan, '</DepartureID>', $iPos1);
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['departure_id'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['departure_position'] = '';

  $iPos1 = strpos ($sCustomFlightPlan, '<DestinationID>') + strlen ('<DestinationID>');
  $iPos2 = strpos ($sCustomFlightPlan, '</DestinationID>', $iPos1);
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['destination_id'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  $iPos1 = strpos ($sCustomFlightPlan, '<DepartureName>') + strlen ('<DepartureName>');
  $iPos2 = strpos ($sCustomFlightPlan, '</DepartureName>', $iPos1);
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['departure_name'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  $iPos1 = strpos ($sCustomFlightPlan, '<DestinationName>') + strlen ('<DestinationName>');
  $iPos2 = strpos ($sCustomFlightPlan, '</DestinationName>', $iPos1);
  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['destination_name'] = substr ($sCustomFlightPlan, $iPos1, $iPos2 - $iPos1);

  $arrNewAtcActiveFlightPlan['ATC_ActiveFlightPlan.0']['alternate_name'] = 'AlternateName';

  $arrCustomFlight = array_merge ($arrBeforeWaypoints, $arrNewAtcActiveFlightPlan, $arrAfterWaypoints);

  if (DEBUG == true)
  {
    /*
    echo '<p>merged Array: $arrCustomFlight</p>';
    print_r ($arrCustomFlight);

    echo '<p>$arrAfterWaypoints</p>';
    print_r ($arrAfterWaypoints);

    echo '<p><p>$arrNewAtcActiveFlightPlan</p>';
    print_r ($arrNewAtcActiveFlightPlan);
    */
  }

  // restore lost 'GPS_Engine'-section from flt-file

  if (!array_key_exists ('GPS_Engine', $arrNewFlightFile) && !array_key_exists ('GPS_Engine', $arrCustomFlight))
  {
    writeMessage ('no GPS_Engine-section found in ' . basename ($sTargetFltFile));
    return true;
  }
  else
  {
    $bSplitpoint = false;
    $arrBeforeCustomFlight = array ();
    $arrAfterCustomFlight = array ();

    foreach ($arrCustomFlight as $key => $val)
    {
      if (stripos ($key, 'DateTimeSeason') === 0)
      {
        if ($bSplitpoint == false)
        {
          $bSplitpoint = true;
          $arrAfterWaypoints[$key] = $val;
        }
      }
      else
      {
        if ($bSplitpoint == false)
          $arrBeforeWaypoints[$key] = $val;
        else
          $arrAfterWaypoints[$key] = $val;
      }
    }
  }

  // missing gps-engine-section in new-flight-file will be replaced by the one found in customflight.flt
  if (!array_key_exists ('GPS_Engine', $arrNewFlightFile) && array_key_exists ('GPS_Engine', $arrCustomFlight))
  {
    $arrCustomFlight = array_merge ($arrBeforeWaypoints, array ('GPS_Engine' => $arrCustomFlight['GPS_Engine']), $arrAfterWaypoints);

    // a missing 'filename'-element should be placed at the beginning of 'GPS_Engine'-section
    if (!isset ($arrCustomFlight['GPS_Engine']['Filename']))
      $arrCustomFlight['GPS_Engine'] = array_merge (array ('Filename' => CUSTOM_FLIGHT_DIR . 'CUSTOMFLIGHT'), $arrNewFlightFile['GPS_Engine']);

    // how many GPS-Waypoints could be found?

    foreach ($arrCustomFlight['GPS_Engine'] as $key => $val)
    {
      if (strpos ($key, 'WpInfo') === 0)
        $iWP++;
    }

    $iNumberOfRestoredGpsWaypoints = $iWP;

    return true;
  }
  elseif (!array_key_exists ('GPS_Engine', $arrNewFlightFile) && !array_key_exists ('GPS_Engine', $arrCustomFlight))
  {
    criticalError ('no GPS_Engine neither in $arrNewFlightFile and $arrCustomFlight: ' . __FUNCTION__ . '()');
    return false;
  }

  // normaly the GPS_Engine'-section should exist in $arrNewFlightFile

  return true;
}


function writeFlightFile ($s, $sTargetFltFile)
{
  // write file according to php.net

  if (DEBUG == true)
    return;

  if (is_writable ($sTargetFltFile))
  {
    if (!$fp = fopen ($sTargetFltFile, 'w'))
    {
      echo "Cannot open file ($sTargetFltFile)";
      exit;
    }

    if (fwrite ($fp, $s) === false)
    {
      echo "Cannot write to file ($sTargetFltFile)";
      exit;
    }

    writeMessage ('writing ' . basename ($sTargetFltFile));

    fclose($fp);

  }
  else
    criticalError ("couldn't write to: $sTargetFltFile");
}


function getStringFromFlightFileArray ($arrFltFile)
{
  $s = '';

  foreach ($arrFltFile as $key => $arr)
  {
    $s .= '[' . $key . ']' . PHP_EOL;

    foreach ($arr as $key2 => $val2)
    {
      $s .= $key2 . '=' . $val2 . PHP_EOL;
    }

    $s .= PHP_EOL;
  }

    return $s;
}


function writeMessage ($s, $bInit = false)
{
  if ($bInit == true)
    $s = PHP_EOL . basename (__FILE__) . ': ' . $s . PHP_EOL;
  else
    $s = '- ' . $s . PHP_EOL;

  if (php_sapi_name () == 'cli')
    echo $s;
  else
    echo '<p><pre style="color:#00aa00;">' . $s . '</pre></p>';
}


function criticalError ($s)
{
  $s = '- CRITICAL ERROR: ' . $s . PHP_EOL;

  if (php_sapi_name () == 'cli')
    echo $s;
  else
    echo '<pre style="color:#ff0000;">' . $s . '</pre>';

  exit;
}


function getUserinput ($s)
{
  echo $s;

  if (php_sapi_name () == 'cli')
  {
    $handle = fopen ("php://stdin","r");

    do
    {
      $line = fgets($handle);
    }
    while ($line == '');

    fclose($handle);

    return trim ($line);
  }
  else
    echo '<p>no action in function ' . __FUNCTION__  . '() when being in ' . php_sapi_name () . ' environment</p>';
}


function switchAircraft ()
{
  global $arrFltFiles, $arrCustomFlight;

  $i = 0;
  $s = '';
  $sNewAircraft = '';
  $sActiveAircraft = '';
  $sSwitchToAircraft = '';
  $sTargetFltFile = FLIGHT_DATA_DIR . getNameOfNewFlightFile (false);

  $bEmptySection = false;

  $arr = array ();
  $arrAircraft = array ();

  // read ini-file with given aircraft names
  $arr = explode (PHP_EOL, trim (file_get_contents (SWITCH_AIRCRAFT_INI)));

  // show a flt-menu in case of temporary flight to choose another $sTargetFltFile

  if (strpos ($sTargetFltFile, TEMPORARY_FLIGHT_FILE))
  {
    if (count ($arrFltFiles) == 0)
    {
      writeMessage ('no alternative flt-files found');
      exit;
    }

    // create menu string for choosing a flt-file
    foreach ($arrFltFiles as $key => $val)
    {
      if ($i > 0)
      {
        $s .= str_pad ($i, 2, " ", STR_PAD_LEFT) . ' - ' . $val . PHP_EOL;
        $arrFltFiles[$i] = $val;
      }
      $i++;
    }

    // show menu and wait for user input
    $s = PHP_EOL . "choose an flt-file (instead of '" . TEMPORARY_FLIGHT_FILE . "')" . PHP_EOL . PHP_EOL . $s;
    $s = getUserinput ($s);

    if (array_key_exists ($s, $arrFltFiles))
      $sTargetFltFile = $arrFltFiles[$s];
    else
      exit;

    writeMessage ("switching to flt-file '" . $sTargetFltFile . "'");
    $sTargetFltFile = FLIGHT_DATA_DIR . $sTargetFltFile;
  }

  // read flt-File (recently saved or choosen) into an array
  $arrTargetFltFile = parse_ini_file ($sTargetFltFile, true, INI_SCANNER_RAW);

  // look into $arrTargetFltFile and check possibility for choosing another aircraft

  if (array_key_exists ('Sim.0', $arrTargetFltFile))
  {
    if (array_key_exists ('Sim', $arrTargetFltFile['Sim.0']))
    {
      // renumber menu options without active aircraft and build $arrTargetFltFile

      $i = 0;
      $s = '';
      $sActiveAircraft = $arrTargetFltFile['Sim.0']['Sim'];

      foreach ($arr as $key => $val)
      {
        if ($val != $sActiveAircraft)
        {
          $i++;
          $arrAircraft[$i] = $val;
        }
      }

      // create menu string for choosing an aircraft

      foreach ($arrAircraft as $key => $val)
      {
        $s .= str_pad ($key, 2, " ", STR_PAD_LEFT) . ' - ' . $val . PHP_EOL;
      }

      $s = PHP_EOL . 'choose an aircraft for: ' . basename ($sTargetFltFile) . " (replacing '" . $sActiveAircraft . "')" . PHP_EOL . PHP_EOL . $s;

      // show menu and wait for user input
      $sSwitchToAircraft = getUserinput ($s);

      // process with an existing key-number
      if (array_key_exists ($sSwitchToAircraft, $arrAircraft))
      {
        $sNewAircraft = $arrAircraft[$sSwitchToAircraft];
        backupFiles ($sTargetFltFile, false, false);
        
        // read recently saved flt-File into an array
        $arrTargetFltFile = parse_ini_file ($sTargetFltFile, true, INI_SCANNER_RAW);

        // replace aircraft in $arrTargetFltFile
        writeMessage ("switching from '" . $sActiveAircraft . "' to '" . $sNewAircraft . "'");
        $arrTargetFltFile['Sim.0']['Sim'] = $sNewAircraft;

        // empty some $arrTargetFltFile-sections
        foreach ($arrTargetFltFile as $key => $val)
        {
          if ($key == 'Freeze.0')
            $bEmptySection = true;
          elseif ($bEmptySection == true)
            $arrTargetFltFile[$key] = [];

          if ($key == 'Controls.0')   // evtl. ab 'Payload.0', die "Wirkung" greift erst für die nachfolgende Sektion (Controls.0 oder Payload.0)
            $bEmptySection = false;
        }

        // overwrite recent flight file
        writeFlightFile (getStringFromFlightFileArray ($arrTargetFltFile), $sTargetFltFile);

        // overwrite customflight.flt file - so sinkrate-ingame-panel may recognize the switched aircraft
        
        // $arrCustomFlight has been modified for several reasons, but will not been saved as customflight.pln, so the file has to be read again
        $arrCustomFlight = parse_ini_file (CUSTOM_FLIGHT, true, INI_SCANNER_RAW);
        $arrCustomFlight['Sim.0']['Sim'] = $sNewAircraft;
        backupFiles ('', false, true);
        writeFlightFile (getStringFromFlightFileArray ($arrCustomFlight), CUSTOM_FLIGHT);
      }
      else
        writeMessage ('no aircraft choosen');
    }
    else
      writeMessage ('no Sim-element');
  }
  else
    writeMessage ('no Sim.0-section');
}


function switchCrashOptions ()
{
  $s = '';
  $arrOptions = array (0 => "don't do anything", 1 => 'set FirstFlightState to PREFLIGHT_HOLDSHORT', 2 => 'set FirstFlightState to FLIGHT_RUNWAY');

  // create menu string for choosing a crash-option
  foreach ($arrOptions as $key => $val)
  {
    $s .= $key . ' - ' . $val . PHP_EOL;
  }

  // show menu and wait for user input
  $s = PHP_EOL . "choose a crash option:" . PHP_EOL . PHP_EOL . $s;

  $s = getUserinput ($s);

  // action according to user input
  switch ($s)
  {
    case '1':
      switchCrashOptionFltFile ('FreeFlight', 'FirstFlightState', 'PREFLIGHT_HOLDSHORT');
      break;
    case '2':
      switchCrashOptionFltFile ('FreeFlight', 'FirstFlightState', 'FLIGHT_RUNWAY');
      break;
    default:
      writeMessage ('exiting with no action');
  }

  exit;
}


function switchCrashOptionFltFile ($sSection, $sKey, $sValue)
{
  $sTargetFltFile = FLIGHT_DATA_DIR . getNameOfNewFlightFile (false, true);
  $arrNewFlightFile = parse_ini_file ($sTargetFltFile, true, INI_SCANNER_RAW);

  writeMessage ('setting [' . $sSection . '][' . $sKey . '] to ' . $sValue . PHP_EOL . '  in ' . "'" . $sTargetFltFile . "'");

  // replace key of the given section with given value
  $arrNewFlightFile[$sSection][$sKey] = $sValue;

  // overwrite recent flight file
  writeFlightFile (getStringFromFlightFileArray ($arrNewFlightFile), $sTargetFltFile);

  // delete lock-file
  $bDeleteLockFile = unlink (LOCK_FILE);

  // show success of delete
  if ($bDeleteLockFile)
    writeMessage ('lock-file deleted');
  else
    criticalError ("couldn't delete '" . LOCK_FILE . "'");
}


//  GEOLOCATION FUNCTIONS


function getDecimalFromCoordinateString ($sCoordinateString)
{
  $arrCoordinates = getArrayFromCoordinateString ($sCoordinateString);

  $decimalCoordinate = DegreesMinutesSeconds2Decimal ($arrCoordinates[0], $arrCoordinates[1], $arrCoordinates[2]);

  return $decimalCoordinate;
}


function getArrayFromCoordinateString ($sCoordinate)
{
  // Expected is something like N48° 56' 9.41" or E8° 58' 31.94".
  // This is given in [SimVars.0]-Latitude and -Longitude.
  // But coordinates of waypoint-data are given with decimal-minutes without seconds - then the decimal part is multiplied by 60 to get the seconds.

  $sSign = '';
  $sCoordinate = trim ($sCoordinate);

  $arr = explode (SEPARATOR_BLANK, $sCoordinate);

  // get signed value from degree-string, because south latitudes and west longitudes are expected as negative values
  $sCompassOrientation = strtolower (substr ($arr[0], 0, 1));

  $arr[0] = substr ($arr[0], 1);

  if ($sCompassOrientation == 's' || $sCompassOrientation == 'w')
    $sSign = SEPARATOR_MINUS_SIGN;

  // clean degree-string from trailing bullshit-sign
  $arr[0] = $sSign . substr ($arr[0], 0, strlen ($arr[0]) - 1);

  // with decimal-only-minutes the decimal part is transformed into seconds and assigned to the missing $arr[2]

  if (!isset ($arr[2]))
  {
    $arrDecimalMinutes = explode (SEPARATOR_PERIOD, $arr[1]);

    $arr[1] = $arrDecimalMinutes[0] . SEPARATOR_MINUTES;
    $arr[2] = (floatval ('0.' . $arrDecimalMinutes[1]) * 60) . SEPARATOR_SECONDS;
  }

  // proceed with #1 and #2 of a 3-entry-array

  // clean minutes
  $arr[1] = substr ($arr[1], 0, strpos ($arr[1], SEPARATOR_MINUTES));

  // clean seconds
  $arr[2] = substr ($arr[2], 0, strpos ($arr[2], SEPARATOR_SECONDS));

  return $arr;
}


function DegreesMinutesSeconds2Decimal ($degrees, $minutes, $seconds)
{
  // inspired by https://www.web-max.ca/PHP/misc_6.php

  $bNegative = false;
  $decCoordinate = 0;
  $degrees = floatval ($degrees);

  // negative degree-values (being south or west) will lead to wrong results when adding the calculated values of minutes and seconds

  // so a negative degree-state is remembered and the degree-value is changed into a positive one
  if ($degrees < 0)
  {
    $bNegative = true;
    $degrees = -$degrees;
  }

//echo "<p>DegreesMinutesSeconds2Decimal() degrees: '" . $degrees . "' '" . $bNegative . "'</p>";


  $minutes = floatval ($minutes);
//echo "<p>DegreesMinutesSeconds2Decimal() minutes: '" . $minutes . "'</p>";
  $seconds = floatval ($seconds);
//echo "<p>DegreesMinutesSeconds2Decimal() seconds: '" . $seconds . "'</p>";

  // with positive degree-values there won't be problems when adding the calculated values of minutes and seconds
  $decCoordinate = $degrees + ((($minutes * 60) + ($seconds)) / 3600);

//echo "<p>DegreesMinutesSeconds2Decimal() decCoordinate: '" . $decCoordinate . "'</p>";

  // a negative degree-state is restored and the whole result-value is changed into a negative one
  if ($bNegative == true)
    $decCoordinate = -$decCoordinate;

  return $decCoordinate;
}


function Decimal2DegreesMinutesSeconds ($decimal)
{
  // inspired by https://www.web-max.ca/PHP/misc_6.php

  // extract the integer part and the float part by a string function to avoid issues with floating point math

  $vars = explode (".", $decimal);
  $degrees = $vars[0];
  $temp = "0." . $vars[1];

  $temp = $temp * 3600;
  $minutes = floor ($temp / 60);
  $seconds = $temp - ($minutes * 60);

  return array ("deg"=>$degrees, "min"=>$minutes, "sec"=>$seconds);
}


function getDistanceBetweenCoordinates ($LatitudeFrom, $LongitudeFrom, $LatitudeTo, $LongitudeTo, $unit = 'nm')
{

  // IMPORTANT: south latitudes and west longitudes are expected with leading '-' as negative values

  // inspired by https://www.geodatasource.com/developers/php

  if (($LatitudeFrom == $LatitudeTo) && ($LongitudeFrom == $LongitudeTo))
    return 0;
  else
  {
    $theta = $LongitudeFrom - $LongitudeTo;
    $dist  = sin (deg2rad ($LatitudeFrom)) * sin (deg2rad ($LatitudeTo)) +  cos (deg2rad ($LatitudeFrom)) * cos (deg2rad ($LatitudeTo)) * cos (deg2rad ($theta));
    $dist  = acos ($dist);
    $dist  = rad2deg ($dist);
    $miles = $dist * 60 * 1.1515;

    if ($unit == "km")
      return ($miles * 1.609344);
    else if ($unit == "nm")
      return ($miles * 0.8684);
    else if ($unit == "m")
      return $miles;
  }
}


function getAirportDataFromLnmApi ($sICAO)
{
  $arr = array ();

  if (USE_LITTLENAVMAP_API == true)
  {
    writeMessage ('connecting LittleNavMap-API-Interface');

    $AssociativeJsonData = @file_get_contents (URL_LNM_API . $sICAO, true);

    if ($AssociativeJsonData == false)
      writeMessage ('LittleNavMap-API-Interface not found (' . $sICAO . ')');
    else
    {
      $AssociativeJsonData =  json_decode ($AssociativeJsonData, 2);    // only with depth-parameter we get standard-array-data

      $arrCoordinates = $AssociativeJsonData['position'];
      $sLongestRunway = $AssociativeJsonData['longestRunwayLength'];

      $arr = array ('lat' => $arrCoordinates['lat'], 'lon' => $arrCoordinates['lon'], 'longestRunwayLength' => $sLongestRunway);
    }
  }
  else
    writeMessage ('continue with ' . $sICAO);

  return $arr;
}

?>